package org.bitbucket.bradleysmithllc.etlunit;

import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.github.fge.jackson.JsonLoader;
import com.github.fge.jsonschema.exceptions.ProcessingException;
import com.github.fge.jsonschema.main.JsonSchema;
import com.github.fge.jsonschema.main.JsonSchemaFactory;
import com.github.fge.jsonschema.report.ProcessingReport;
import org.bitbucket.bradleysmithllc.etlunit.feature.Feature;
import org.bitbucket.bradleysmithllc.etlunit.feature.RuntimeOption;
import org.bitbucket.bradleysmithllc.etlunit.io.ApacheProcessStreamsExecutor;
import org.bitbucket.bradleysmithllc.etlunit.io.FileBuilder;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestClass;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestMethod;
import org.bitbucket.bradleysmithllc.etlunit.parser.ETLTestValueObject;
import org.bitbucket.bradleysmithllc.etlunit.regexp.PackNameExpression;
import org.bitbucket.bradleysmithllc.etlunit.util.*;

import javax.inject.Inject;
import javax.inject.Named;
import java.io.*;
import java.net.InetAddress;
import java.net.NetworkInterface;
import java.security.MessageDigest;
import java.util.*;

public class BasicRuntimeSupport implements RuntimeSupport {
	private final File root;
	private final File tempDir;
	private final File resourceDir;
	private final File referenceDir;
	private final File configDir;
	private final File generatedSourcesDir;
	private final File reportRootDir;

	private MapLocal mapLocal;

	private Log applicationLog;
	private Log userLog;

	private ProcessExecutor processExecutor;
	private final File srcRoot;
	private final String projectName;
	private final String projectVersion;
	private final String projectUser;
	private final File vendorBinaryDir;
	private final String projectUid;

	private final List<Feature> featureList;

	private final Map<String, RuntimeOption> runtimeOptions = new HashMap<String, RuntimeOption>();

	private static final JsonSchema configurationJsonSchema;

	static {
		JsonSchemaFactory fact = JsonSchemaFactory.byDefault();

		try {
			JsonNode node = JsonSchemaResolver.resolveSchemaReference(JsonLoader.fromURL(RuntimeSupport.class.getClassLoader().getResource("org/bitbucket/bradleysmithllc/etlunit/Configuration.jsonSchema")));

			JsonSchemaValidator.validateV4Schema(node);

			configurationJsonSchema = fact.getJsonSchema(node);
		} catch (ProcessingException e) {
			throw new RuntimeException("Could not process operation schema: " + JsonProcessingUtil.getProcessingReport(e), e);
		} catch (Exception e) {
			throw new RuntimeException("Could not load operation schema", e);
		}
	}

	public BasicRuntimeSupport() {
		this(getSystemTempDir());
	}

	public BasicRuntimeSupport(File proot) {
		featureList = Collections.EMPTY_LIST;

		try {
			this.root = proot.getCanonicalFile();
		} catch (IOException e) {
			throw new RuntimeException(e);
		}

		this.tempDir = new File(root, "temp");
		tempDir.mkdirs();

		this.reportRootDir = new File(root, "reports");
		reportRootDir.mkdirs();

		vendorBinaryDir = tempDir;

		this.resourceDir = new File(root, "resource");
		resourceDir.mkdirs();

		this.referenceDir = new File(root, "reference");
		referenceDir.mkdirs();

		configDir = new File(resourceDir, "config");
		configDir.mkdirs();

		generatedSourcesDir = new File(root, "build");
		generatedSourcesDir.mkdirs();

		projectName = "proj";
		projectVersion = "1_0";
		projectUser = "builduser";

		// use the apache commons exec implementation by default
		installProcessExecutor(new ApacheProcessStreamsExecutor());
		srcRoot = new File(this.root, "src");
		projectUid = readProjectUid(System.getProperty("project.uid")).toLowerCase();
	}

/*
	public BasicRuntimeSupport(File root, File src, File temp, File rsrc, File cfg) {
		this.root = root;
		this.temp = temp;
		this.srcRoot = src;
		resourceDir = rsrc;
		configDir = cfg;

		// use the apache commons exec implementation by default
		installProcessExecutor(new ApacheCommonsStreamsExecutor());
	}
	 */

	public BasicRuntimeSupport(Configuration configuration, ETLTestVM vm) {
		featureList = vm.getFeatures();

		// validate against the schema
		JsonNode configJson = configuration.getRoot().getJsonNode();
		try {
			JsonSchemaFactory fact = JsonSchemaFactory.byDefault();

			ProcessingReport report = configurationJsonSchema.validate(configJson);

			if (!report.isSuccess()) {
				throw new RuntimeException("Configuration schema does not validate: " + JsonProcessingUtil.getProcessingReport(report));
			}
		} catch (ProcessingException e) {
			throw new RuntimeException("Configuration schema does not validate: " + JsonProcessingUtil.getProcessingReport(e), e);
		}

		// configuration is valid - proceed
		projectName = StringUtils.sanitize(configuration.query("project-name").getValueAsString(), '_').toLowerCase();
		projectVersion = StringUtils.sanitize(configuration.query("project-version").getValueAsString(), '_').toLowerCase();
		projectUser = StringUtils.sanitize(configuration.query("project-user").getValueAsString(), '_').toLowerCase();

		try {
			root = new File(configuration.query("project-root-directory").getValueAsString()).getCanonicalFile();

			srcRoot = new File(configuration.query("test-sources-directory").getValueAsString()).getCanonicalFile();
			configDir = new File(configuration.query("configuration-directory").getValueAsString()).getCanonicalFile();
			generatedSourcesDir =
					new File(configuration.query("generated-source-directory").getValueAsString()).getCanonicalFile();
			resourceDir = new File(configuration.query("resource-directory").getValueAsString()).getCanonicalFile();
			referenceDir = new File(configuration.query("reference-directory").getValueAsString()).getCanonicalFile();
			tempDir = new File(configuration.query("temp-directory").getValueAsString()).getCanonicalFile();
			reportRootDir = new File(configuration.query("reports-directory").getValueAsString()).getCanonicalFile();

			vendorBinaryDir = new File(configuration.query("vendor-binary-directory").getValueAsString()).getCanonicalFile();
		} catch (IOException e) {
			throw new IllegalArgumentException(e);
		}

		if (!tempDir.exists()) {
			tempDir.mkdirs();
		}

		if (!vendorBinaryDir.exists()) {
			vendorBinaryDir.mkdirs();
		}

		installProcessExecutor(new ApacheProcessStreamsExecutor());

		ETLTestValueObject query = configuration.query("project-uid");
		projectUid = readProjectUid(query != null ? query.getValueAsString() : null);

		ETLTestValueObject optionsNode = configuration.query("options");
		if (optionsNode != null) {
			ObjectMapper om = new ObjectMapper();

			Map<String, ETLTestValueObject> map = optionsNode.getValueAsMap();

			for (Map.Entry<String, ETLTestValueObject> featureEntry : map.entrySet()) {
				// at this level read the actual properties
				String featureNasme = featureEntry.getKey();
				ETLTestValueObject featureObject = featureEntry.getValue();

				Map<String, ETLTestValueObject> featureMap = featureObject.getValueAsMap();

				for (Map.Entry<String, ETLTestValueObject> optionEntry : featureMap.entrySet()) {
					// at this level read the actual properties
					String optionName = optionEntry.getKey();
					ETLTestValueObject optionObject = optionEntry.getValue();

					if (optionObject != null) {
						JsonNode optionObjectJsonNode = optionObject.getJsonNode();

						if (!optionObjectJsonNode.isNull())
						{
							String json = optionObjectJsonNode.toString();

							try {
								RuntimeOption runtimeOption = om.readValue(json, RuntimeOption.class);
								runtimeOption.setName(featureNasme + "." + optionName);
								runtimeOption.validate();

								runtimeOptions.put(runtimeOption.getName(), runtimeOption);
							} catch (IOException e) {
								throw new ExceptionInInitializerError(e);
							}
						}
					}
				}
			}
		}
	}

	private String readProjectUid(String uid) {
		if (uid != null) {
			return uid.toLowerCase();
		}

		try {
			StringBuffer
					buffer =
					new StringBuffer(root.getAbsolutePath()).append("|")
							.append(projectName)
							.append("|")
							.append(projectVersion)
							.append("|");

			InetAddress inet = InetAddress.getLocalHost();

			if (!inet.isLoopbackAddress()) {
				NetworkInterface ni = NetworkInterface.getByInetAddress(inet);

				String mac = Hex.hexEncode(ni.getHardwareAddress());

				buffer.append(mac);
			} else {
				buffer.append(inet.getCanonicalHostName());
			}

			byte[] bytesOfMessage = buffer.toString().getBytes("UTF-8");

			MessageDigest md = MessageDigest.getInstance("MD5");
			String thedigest = Hex.hexEncode(md.digest(bytesOfMessage));

			return thedigest.toLowerCase();
		} catch (Exception e) {
			throw new RuntimeException("Error reading local host information", e);
		}
	}

	@Inject
	public void setApplicationLogger(@Named("applicationLog") Log log) {
		this.applicationLog = log;
	}

	@Inject
	public void setUserLogger(@Named("userLog") Log log) {
		this.userLog = log;
	}

	public static File getSystemTempDir() {
		return new File(System.getProperty("java.io.tmpdir"));
	}

	@Override
	public String getProjectRelativePath(File fke) throws IOException {
		// determine if the path parameter is truly a subpath of the project root
		String absolutePath = fke.getCanonicalPath();
		String rootAbsolutePath = root.getAbsolutePath();

		// special deal - if the path IS the project root, return '.'
		if (absolutePath.equals(rootAbsolutePath))
		{
			return ".";
		}

		if (!absolutePath.startsWith(rootAbsolutePath))
		{
			throw new IllegalArgumentException("Path is not a subpath of the project root");
		}

		String subPath = absolutePath.substring(rootAbsolutePath.length());

		// unixify the path
		subPath = subPath.replace(File.separator, "/");

		// make sure that this path fragment does not begin with a slash
		if (subPath.startsWith("/"))
		{
			subPath = subPath.substring(1);
		}

		// we don't want it to end with one either
		if (subPath.endsWith("/"))
		{
			subPath = subPath.substring(0, subPath.length() - 1);
		}

		return subPath;
	}

	@Override
	public File getTempDirectory() {
		return tempDir;
	}

	@Override
	public Log getApplicationLog() {
		return applicationLog;
	}

	@Override
	public Log getUserLog() {
		return userLog;
	}

	@Override
	public File createTempFile(String name) {
		return new File(getTempDirectory(), name);
	}

	@Override
	public File getGeneratedSourceDirectory(String feature) {
		return new FileBuilder(generatedSourcesDir).subdir(feature).mkdirs().file();
	}

	@Override
	public File createGeneratedSourceFile(String feature, String name) {
		return new File(getGeneratedSourceDirectory(feature), name);
	}

	@Override
	public File getFeatureSourceDirectory(String feature) {
		return new FileBuilder(resourceDir).subdir(feature).mkdirs().file();
	}

	@Override
	public File getSourceDirectory() {
		return srcRoot;
	}

	@Override
	public File getTestSourceDirectory() {
		return getTestSourceDirectory(null);
	}

	@Override
	public File getTestSourceDirectory(String _package) {
		FileBuilder fb = new FileBuilder(getSourceDirectory());

		if (_package != null && !_package.equals("")) {
			// split so deep packages can be handled
			String[] packageNames = _package.split("\\.");

			for (String pack : packageNames) {
				fb = fb.subdir("_" + pack + "_");
			}
		}

		return fb.file();
	}

	@Override
	public File getCurrentTestSourceDirectory() {
		return getTestSourceDirectory(getCurrentlyProcessingTestPackage());
	}

	@Override
	public String getCurrentlyProcessingTestPackage() {
		return getCurrentlyProcessingTestClass().getPackage();
	}

	@Override
	public ETLTestClass getCurrentlyProcessingTestClass() {
		return mapLocal.getCurrentlyProcessingTestClass();
	}

	@Override
	public ETLTestMethod getCurrentlyProcessingTestMethod() {
		return mapLocal.getCurrentlyProcessingTestMethod();
	}

	@Override
	public File getTestResourceDirectory(String subdir) {
		return getTestResourceDirectory(null, subdir);
	}

	@Override
	public File getTestResourceDirectory(String _package, String subdir) {
		File packRoot = getTestSourceDirectory(_package);

		return new FileBuilder(packRoot).subdir(subdir).file();
	}

	@Override
	public File getCurrentTestResourceDirectory(String subdir) {
		return getTestResourceDirectory(getCurrentlyProcessingTestClass().getPackage(), subdir);
	}

	@Override
	public File getReportDirectory(String type) {
		return new FileBuilder(reportRootDir).subdir(type + "-reports").mkdirs().file();
	}

	@Override
	public String getProjectName() {
		return projectName;
	}

	@Override
	public String getProjectVersion() {
		return projectVersion;
	}

	@Override
	public String getProjectUser() {
		return projectUser;
	}

	@Override
	public String getProjectUID() {
		return projectUid;
	}

	@Override
	public void overrideRuntimeOption(RuntimeOption runtimeOption) {
		runtimeOptions.put(runtimeOption.getName(), runtimeOption);
	}

	@Override
	public void overrideRuntimeOptions(List<RuntimeOption> runtimeOptions) {
		for (RuntimeOption option : runtimeOptions) {
			overrideRuntimeOption(option);
		}
	}

	@Override
	public File getFeatureConfigurationDirectory(String feature) {
		return new FileBuilder(configDir).subdir(feature).mkdirs().file();
	}

	@Override
	public File getSourceDirectory(String package_) {
		if (package_ == null || package_.equals("")) {
			return getSourceDirectory();
		}

		return new FileBuilder(getSourceDirectory()).subdir(package_.replace('.', File.separatorChar)).mkdirs().file();
	}

	@Override
	public ProcessExecutor getProcessExecutor() {
		return processExecutor;
	}

	@Override
	public File createSourceFile(String feature, String name) {
		return new File(getSourceDirectory(feature), name);
	}

	@Override
	public File getReferenceDirectory(String path_) {
		return new FileBuilder(referenceDir).subdir(path_).mkdirs().file();
	}

	@Override
	public File getReferenceFile(String path_, String name) {
		return new File(getReferenceDirectory(path_), name);
	}

	@Incomplete
	@NeedsTest
	@Override
	public void installProcessExecutor(ProcessExecutor od) {
		processExecutor = od;
	}

	@Incomplete
	@NeedsTest
	public ProcessFacade execute(final ProcessDescription pd) throws IOException {
		applicationLog.info("Executing process: " + pd.debugString());

		File fi = pd.getOutputFile();

		if (fi != null) {
			applicationLog.info("Logging process output to file: " + fi.getAbsolutePath());
		}

		if (processExecutor != null) {
			return processExecutor.execute(pd);
		}

		ProcessBuilder pb = new ProcessBuilder(pd.getCommand());
		if (pd.hasArguments()) {
			pb = pb.command(pd.getArguments());
		}

		pb.environment().putAll(pd.getEnvironment());


		final Process proc = pb.start();

		return new ProcessFacade() {
			@Override
			public ProcessDescription getDescriptor() {
				return pd;
			}

			public Process getProcess() {
				return proc;
			}

			@Override
			public void waitForCompletion() {
				try {
					proc.waitFor();
				} catch (InterruptedException e) {
					throw new IllegalStateException("Process interrupted");
				}
			}

			@Override
			public int getCompletionCode() {
				return proc.exitValue();
			}

			@Override
			public void kill() {
				proc.destroy();
			}

			@Override
			public void waitForStreams() {
				//these are always available
			}

			@Override
			public void waitForOutputStreamsToComplete() {
			}

			@Override
			public Writer getWriter() {
				return new BufferedWriter(new OutputStreamWriter(getOutputStream()));
			}

			@Override
			public BufferedReader getReader() {
				return new BufferedReader(new InputStreamReader(getInputStream()));
			}

			@Override
			public BufferedReader getErrorReader() {
				return new BufferedReader(new InputStreamReader(getErrorStream()));
			}

			@Override
			public StringBuffer getInputBuffered() throws IOException {
				return buffer(getReader());
			}

			@Override
			public StringBuffer getErrorBuffered() throws IOException {
				return buffer(getErrorReader());
			}

			private StringBuffer buffer(BufferedReader errorReader) throws IOException {
				StringBuffer stb = new StringBuffer();

				String line = null;

				int lineC = 0;

				while ((line = errorReader.readLine()) != null) {
					if (lineC != 0) {
						stb.append('\n');
					}

					lineC++;

					stb.append(line);
				}

				return stb;
			}

			@Override
			public OutputStream getOutputStream() {
				return proc.getOutputStream();
			}

			@Override
			public InputStream getInputStream() {
				return proc.getInputStream();
			}

			@Override
			public InputStream getErrorStream() {
				return proc.getErrorStream();
			}
		};
	}

	@Override
	public RuntimeOption getRuntimeOption(String qualifiedName) {
		return runtimeOptions.get(qualifiedName);
	}

	@Override
	public List<RuntimeOption> getRuntimeOptions(String featureName) {
		List<RuntimeOption> options = new ArrayList<RuntimeOption>();

		for (Map.Entry<String, RuntimeOption> entry : runtimeOptions.entrySet()) {
			String key = entry.getKey();
			RuntimeOption value = entry.getValue();

			if (featureName == null || key.startsWith(featureName + ".")) {
				options.add(value);
			}
		}

		return options;
	}

	@Override
	public File getVendorBinaryDir(String path) {
		return new FileBuilder(vendorBinaryDir).subdir(path).file();
	}

	@Override
	public File getVendorBinaryDir() {
		return vendorBinaryDir;
	}

	@Override
	public List<RuntimeOption> getRuntimeOptions() {
		return getRuntimeOptions(null);
	}

	private void getPackages(File root, final String prefix, final List<String> packages, final PackageValidator validator) {
		root.listFiles(new FileFilter() {
			public boolean accept(File pathname) {
				if (pathname.isDirectory()) {
					PackNameExpression pne = new PackNameExpression(pathname.getName());

					if (pne.matches()) {
						String pack = (prefix != null ? (prefix + ".") : "") + pne.getPackName();
						packages.add(pack);

						getPackages(pathname, pack, packages, validator);
					}
				}

				return false;
			}
		});
	}

	@Override
	public List<String> getPackages(File root) {
		final List<String> packages = new ArrayList<String>();

		// add a reference to the default package
		packages.add(null);

		getPackages(root, null, packages, null);

		return packages;
	}

	@Override
	public List<String> getReferencePackages(String path) {
		File referenceDirectory = getReferenceDirectory(path);
		return getPackages(referenceDirectory);
	}

	interface PackageValidator {
		boolean packageNameValid(String name);
	}

	@Override
	public List<String> getTestPackages() {
		final List<String> packages = new ArrayList<String>();

		// add a reference to the default package
		packages.add(null);

		getPackages(getTestSourceDirectory(), null, packages, new TestPackageValidator());

		return packages;
	}

	private class TestPackageValidator implements PackageValidator {
		private final Map<String, String> nameMap = new HashMap<String, String>();

		TestPackageValidator() {
			for (Feature feature : featureList) {
				for (String dirName : feature.getTestSupportFolderNames()) {
					nameMap.put(dirName, dirName);
				}
			}
		}

		@Override
		public boolean packageNameValid(String name) {
			return !nameMap.containsKey(name);
		}
	}

	public void setMapLocal(MapLocal mapLocal) {
		this.mapLocal = mapLocal;
	}
}